// module waveClass3
// Copyright Dr.E.Huckert 07-2011
// Plays a liste of wave files
// Only for Windows (MultiMedia Library winmm.lib)
// This version uses the more complex API calls: waveOutOpen(), waveOutWrite() etc.
// Uses a separate thread for each wave file to be played
// A callback function is used to detect the end of the playing process (WOM_DONE)
// Signals (=events) are sent to indicate the termination of the  threads

// Compile with (Digital Mars C++):
//  dmc -mn -WA -Ae waveClass3.cpp winmm.lib
// For Microsoft: thread safe libraries must be used!

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <mmsystem.h>

#define MAX_NO_FILES 16

class      WaveFile; 
static int logging = 0; 

// this class is used by the player threads
class ThreadParams
{
  public:
  char   actFn[256];
  HANDLE event;
};

// classes (structures) used for reading the WAVE files
class MAIN_CHUNK
{
  public:
  DWORD main_chunk;  // must be "RIFF"
  DWORD mlength;     // Laenge ueber alles ohne die ersten 8 Bytes
  DWORD chunk_type;  // must be "WAVE"
};        // 12 bytes

class FORMAT_CHUNK     // start byte #12
{
  public:
  DWORD sub_chunk;   // muss 'fmt ' sein
  DWORD slength;     // #16
  WORD  format;      // #18
  WORD  channels;    // #20  1=mono, 2=stereo
  DWORD sample_fq;   // #24
  DWORD bytes_per_second;
  WORD  bytes_per_sample;
  WORD  bits_per_sample;
};        // 24 bytes

class DATA_CHUNK     // start at pos. (12 + 24)=36
{
  public:
  DWORD data_type;   // kann 'data' sein oder 'LINK' oder ...
  DWORD data_length;
};      // 8 bytes

// ----------------------------------------------------
// A class to evaluate and play WAVE files.
// class written by EH 01-2015
// modified/enhanced  11-2015
class WaveFile
{
  public:
    int           no_channels;
    int           bits_samp;
    int           bytes_samp;
    unsigned long sample_rate;     
    unsigned long bytes_sec;
    unsigned long length;
    WAVEFORMATEX  waveform;
    PWAVEHDR      pWaveHdr;
    char          fileName[256];
    DWORD         callBckParam;
    HWAVEOUT      hWaveOut;
    unsigned char *pData;
    //
  public:
    WaveFile(const char *file_name);  // constructor
    ~WaveFile();
    int readMainChunk(int ifd);
    int readFormatChunk(int ifd);
    long readDataChunks(int ifd);
    int readAttributes(int dataFlag); 
    int play(HANDLE event);           // use WaveOut() to play
    int close();
    int open();
};

// -----------------------------------------------------------------
// This function is called by the Windows waveOutXXX() functions when
// certain states are reached (like MM_WOM_DONE when the play is over)
// The message type is the returned in the pInstance parameter. This parameter
// corresponds to the callBckParam variable in class WaveFile (which is thread specific)
void CALLBACK CallBckProcedure(HWAVEOUT  hwo, 
                               UINT      message, 
                               DWORD     *pInstance, // return data
                               DWORD     *pParam1,
                               DWORD     *param2)
{
  int ret = 0;
  //
  switch (message)  
  {
    case MM_WOM_OPEN:
      if (::logging)
        ::printf("CallBckProcedure WOM_OPEN\n");
      ret = 1;
      break;

    case MM_WOM_DONE:
      if (::logging)
        ::printf("CallBckProcedure WOM_DONE\n");
      ret = 2;
      break;

    case MM_WOM_CLOSE:
      if (::logging)
        ::printf("CallBckProcedure WOM_CLOSE\n");
      ret = 3;
      break;

    default: 
      ret = 4;
      break; // do nothing for all other messages
  }   // end switch (message)
  if (pInstance != NULL)
    *pInstance = ret;
  if (::logging)
    ::printf("CallBckProcedure ret=%d\n", ret);
}   // end CallBckProcedure()

// -------------------------------------------
// constructor with a given file name(WAV file)
WaveFile::WaveFile(const char *file_name)
{
  if (file_name == NULL)
    ::strcpy(this->fileName, "");
  else
    ::strcpy(this->fileName, file_name);
  this->no_channels  = 1;
  this->bytes_samp   = 0;
  this->bits_samp    = 0;
  this->length      = 0L;
  this->bytes_sec   = 0L;
  this->sample_rate = 0L;
  this->pWaveHdr    = (PWAVEHDR)new char[sizeof(WAVEHDR)];
  this->callBckParam = 0;
  this->hWaveOut    = 0;
  this->pData       = NULL;  
}

// ----------------------------------------------------
// Destructor
WaveFile::~WaveFile()
{
  if (this->hWaveOut != 0)
    this->close();
  if (this->pWaveHdr != NULL)
    delete this->pWaveHdr; 
  if (this->pData != NULL)
    delete this->pData;
}

// ------------------------------------------------
// closes the sound device
int WaveFile::close()
{
  int ret = 0;
  MMRESULT mmRet;
  //
  if (this->hWaveOut != 0)
  {
    mmRet = ::waveOutClose(this->hWaveOut);
    if (mmRet != MMSYSERR_NOERROR)
    {
      char msg[256];
      ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
      if (::logging)
      printf("WaveFile::waveOutClose() reports ERROR: %s\n", msg);
      ret = -1;
    }
    this->hWaveOut = 0;
  }
  if (::logging)
      printf("WaveFile::waveOutClose() ret=%d\n", ret);
  return ret;
}   // end WaveFile::close()

// ---------------------------------------------------
// Opens the sound output device
// For multiple simultaneous sounds this must be called for each 
// WAVE file - don't close the device too early!
// Returns < 0 upon error
int WaveFile::open()
{
  int ret = 0;
  MMRESULT mmRet;
  //
  // Note: the WAVE device is opened once. 
  //            This means that all files to be played must have the same essential attributes!!! 
  //             The code could also be placed  in a static method
  this->waveform.wFormatTag      = WAVE_FORMAT_PCM;
  this->waveform.nChannels       = this->no_channels;   
  this->waveform.nSamplesPerSec  = this->sample_rate;  
  this->waveform.nAvgBytesPerSec = this->bytes_sec; 
  this->waveform.wBitsPerSample  = this->bits_samp; 
  this->waveform.nBlockAlign     = (this->bits_samp >> 3) * 
                                    this->no_channels;
  this->waveform.cbSize          = 0;  // extra bytes not needed
  if (this->hWaveOut == 0)
  {
    mmRet = ::waveOutOpen(&(this->hWaveOut), 
                          WAVE_MAPPER, 
                          (WAVEFORMATEX *)(&(this->waveform)),
                          (unsigned long)CallBckProcedure, 
                           // important: the API will return in callBckParam the last message type
                          (unsigned long)(&(this->callBckParam)), 
                          CALLBACK_FUNCTION); 
    if (mmRet != MMSYSERR_NOERROR)
    {
      char msg[256];
      ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
      printf("WaveFile::play() waveOutOpen() reports ERROR: %s\n", msg);
      ret = -1;
      goto zurueck;
    }
    if (::logging)
      printf("WaveFile::open() waveOutOpen() finished\n");
    // not sure whether this is needed 
    mmRet = ::waveOutReset(this->hWaveOut);
    if (mmRet != MMSYSERR_NOERROR)
    {
      char msg[128];
      ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
      printf("WaveFile::open() waveOutReset() reports ERROR: %s\n", msg);
      ret = -2;
      goto zurueck;
    }
    if (::logging)
      printf("WaveFile::open() waveOutReset() finished\n");
    ret = 1;
  }  // if (this->hWaveOut == 0)
  //
  zurueck:
  return ret;  
}   // end WaveFile::open()

// ----------------------------------------------------------
// read 12 bytes  main chunk header (WAVE file)
// Returns: the number of bytes read
//          < 0 if error
int WaveFile::readMainChunk(int ifd)
{
  int nb;
  int ret = 0;
  MAIN_CHUNK header;
  //
  nb = ::read(ifd, &header, sizeof(MAIN_CHUNK));
  if (nb != sizeof(MAIN_CHUNK))
  {
    ret = -1;
    goto zurueck;
  }
  if (::strncmp((char *)&(header.main_chunk), "RIFF", 4) != 0)
  {
    ret = -2;
  }
  if (::strncmp((char *)&(header.chunk_type), "WAVE", 4) != 0)
  {
    ret = -3;
  }
  // everything OK here
  ret = nb;
  //
  zurueck:
  if (::logging)
      printf("WaveFile::readMainChunk() ret=%d\n", ret);
  return ret;
}   // end WaveFile::readMainChunk()

// ----------------------------------------------------------
// read the format chunk  (WAVE file)
// Returns: the number of bytes read
//          < 0 if error
int WaveFile::readFormatChunk(int ifd)
{
  int nb;
  int ret = 0;
  FORMAT_CHUNK format;
  //
  nb = ::read(ifd, &format, sizeof(FORMAT_CHUNK));
  if (nb != sizeof(FORMAT_CHUNK))
  {
    ret = -1;
    goto zurueck;
  }
  if (::strncmp((char *)&(format.sub_chunk), "fmt ", 4) != 0)
  {
    ret = -2;
    goto zurueck;
  }
  // This is a "fmt " sub chunk.   transfer the attributes
  this->no_channels = format.channels;
  this->sample_rate = format.sample_fq;
  this->bytes_samp  = format.bytes_per_sample;
  this->bytes_sec   = format.bytes_per_second;
  this->bits_samp   = format.bits_per_sample;
  // everything OK here
  ret = nb;
  //
  zurueck:
  if (::logging)
      printf("WaveFile::readFormatChunk() ret=%d\n", ret);
  return ret;
}   // end WaveFile::readFormatChunk()

// ----------------------------------------------------------
// Read the data chunks, allocate memory for the data  (WAVE file)
// We evaluate only chunks of type "data" - other chunks are read but
// not evaluated.
// Returns: the number of bytes read
//          < 0 if error
long WaveFile::readDataChunks(int ifd)
{
  long nb;
  long nbTotal = 0L;
  long stored  = 0L;
  long ret = 0;
  DATA_CHUNK data;
  unsigned char *pDest = NULL;
  char buf[1024];
  //
  while (1)
  {
    nb = ::read(ifd, &data, sizeof(DATA_CHUNK));
    if (nb != sizeof(DATA_CHUNK))
    {
      if (nbTotal > 0L)
        break;
      ret = -1;
      goto zurueck;
    }
    nbTotal += nb;
    if (::logging)
    {
      ::memcpy(buf, &(data.data_type), 4);
      buf[4] = 0;
      printf("WaveFile::readDataChunks() type=%s\n", buf);
      printf("WaveFile::readDataChunks() length=%ld\n", data.data_length);
    }
    if (::strncmp((char *)(&(data.data_type)), "data", 4) == 0)
    {
      // note: we reserve space only for the first DATA chunk !!!  - must be changed
      this->length += data.data_length;
      if (this->pData == NULL)
      {
        this->pData = new unsigned char[this->length];
        pDest = this->pData;
      }
    }
    while (data.data_length > 0L)
    {
      nb = sizeof(buf);
      if (nb > data.data_length)
        nb = data.data_length;
      nb = ::read(ifd, buf, nb);  
      if (nb <= 0)
      {
        ret = -2;
        goto zurueck;
      }
      nbTotal += nb;
      data.data_length -= nb;
      if (::strncmp((char *)(&(data.data_type)), "data", 4) == 0)
      {
        if ((pDest != NULL) && ((stored + nb) <= this->length))
        {
           // note: store only the first DATA chunk !!!  - must be changed
          ::memcpy(pDest, buf, nb);
          pDest += nb;
          stored += nb;
        }
      }
    }
  }   // end while (1)
  //
  // everything OK here
  ret = nbTotal;
  //
  zurueck:
  if (::logging)
      printf("WaveFile::readDataChunks() ret=%ld\n", ret);
  return ret;
}   // end WaveFile::readDataChunks()

// ---------------------------------------------------
// Open the file, evaluate the attributes.
// Read and keep the data of flag dataFlag is set
int WaveFile::readAttributes(int dataFlag)
{
  int   ret  = 0; 
  int   cRet = 0;
  long   lRet = 0L;
  int   ifd  = -1;
  FILE  *stream = NULL;
  char  buff[16];
  int   state = 0;
  long  offset;
  //
  ifd = ::open(this->fileName, O_RDONLY + O_BINARY);
  if (ifd < 0) 
  {
    if (::logging)
      printf("WaveFile::readAttributes() ERROR cannot open file=%s\n", this->fileName);
    ret = -1;
    goto zurueck;
  }
  // the  file is here open - read the file header
  cRet = this->readMainChunk(ifd);
  if (cRet < 0)
  {
    // probably wrong file type
    if (::logging)
      printf("WaveFile::readAttributes() ERROR wrong type? file=%s\n", this->fileName);
    ret = -2;
    goto zurueck;
  }
  state = 1;
  //
  // get and keep the essential sound attributes
  cRet = this->readFormatChunk(ifd);
  if (cRet < 0)
  {
    if (::logging)
      printf("WaveFile::readAttributes() ERROR wrong type? file=%s\n", this->fileName);
    ret = -3;
    goto zurueck;
  }
  state = 2;
  if (dataFlag)
  {
    // read and keep the wave data
    state = 3;
    lRet = this->readDataChunks(ifd);
    if (lRet < 0L)
    {
      ret = -4;
      goto zurueck;
    }
  }   // end if (dataFlag)
  //
  // everything OK here
  state = 4;
  if (::logging)
  {
    ::printf("WaveFile::readAttributes() no_channels=%d\n", this->no_channels);
    ::printf("WaveFile::readAttributes() length=%ld\n",     this->length);
    ::printf("WaveFile::readAttributes() sample_rate=%ld\n", this->sample_rate);
    ::printf("WaveFile::readAttributes() bytes_samp=%d\n",  this->bytes_samp);
    ::printf("WaveFile::readAttributes() bytes_sec=%ld\n",  this->bytes_sec);
    ::printf("WaveFile::readAttributes() bits_samp=%d\n",   this->bits_samp);
  }
  //
  zurueck:
  if (ifd >= 0)
    ::close(ifd);
  if (::logging)
    ::printf("WaveFile::readAttributes() ret=%d state=%d\n", 
             ret, state);
  return ret;
}   // end WaveFile::readAttributes()

// ---------------------------------------------------
// Thread worker routine
DWORD WINAPI workerRoutine(LPVOID pParams)
{
  int startPos = 0;
  int ret      = 0;
  int cRet;
  WaveFile *pwf = NULL;
  int noFiles  = 0;
  ThreadParams *ptp = NULL;
  //
  if (::logging)
  {
    if (ptp != NULL)
      printf("Thread worker routine started file name=%s\n", 
            ptp->actFn);
  }
  ptp = (ThreadParams *)pParams;
  if (ptp == NULL)
  {
    ret = -1;
    goto zurueck;
  }
  try
  {
    pwf  = new WaveFile((const char *)ptp->actFn);
    // start the player
    cRet = pwf->play(ptp->event);
    if (cRet < 0)
    {
      ret = -2;
      goto zurueck;
    }
  }   // end try
  catch (...)
  {
    printf("EXCEPTION in Thread routine workerRoutine()");
    ret = -3;
    goto zurueck;
  }
  //
  zurueck:
  if (::logging)  
    printf("thread finished ret=%d\n", ret);
  return ret;
}   // end workerRoutine()

// ----------------------------------------------------------
// Play the file by using API call  ::WaveOut()
int WaveFile::play(HANDLE event)
{
  int      ret = 0;
  int      cRet;
  MMRESULT mmRet;
  DWORD    dword;
  DWORD    dwRepetitions = 1; 
  double   noMs = 0.0;
  //
  if (::logging)
    ::printf("WaveFile::play() file=%s\n", this->fileName);
  if (::strlen(this->fileName) == 0)
  {
    ret = -1;
    goto zurueck;
  }
  //
  // this call must be executed! Evaluates the attributes, reads and keeps the wave data
  cRet = this->readAttributes(1);
  if (cRet < 0)
  {
    ret = -2;
    goto zurueck;
  }
  //
  this->pWaveHdr->lpData          = (char *)(this->pData);
  this->pWaveHdr->dwBufferLength  = this->length; 
  this->pWaveHdr->dwBytesRecorded = 0;
  this->pWaveHdr->dwUser          = 0;
  // if only one buffer ist used both flags must be set
  this->pWaveHdr->dwFlags         = WHDR_BEGINLOOP |
                                    WHDR_ENDLOOP;
  this->pWaveHdr->dwLoops         = dwRepetitions;
  this->pWaveHdr->lpNext          = NULL;
  this->pWaveHdr->reserved        = 0;
  //
  // Note: the audio device must be opend for each WAVE file (each thread)
  if (this->open() < 0)
  {
    ret = -3;
    goto zurueck;
  }
  //
  mmRet = ::waveOutPrepareHeader(this->hWaveOut, 
                                 this->pWaveHdr, 
                                 sizeof(WAVEHDR));
  if (mmRet != MMSYSERR_NOERROR)
  {
    char msg[256];
    ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
    printf("WaveFile::play() waveOutPrepareHeader() reports ERROR: %s\n", msg);
    ret = -4;
    goto zurueck;
  }
  //
  mmRet = ::waveOutWrite(this->hWaveOut, 
                         this->pWaveHdr, 
                         sizeof(WAVEHDR));
  if (mmRet != MMSYSERR_NOERROR)
  {
    char msg[256];
    ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
    printf("WaveFile::play() waveOutWrite() reports ERROR: %s\n", msg);
    ret = -5;
    goto zurueck;
  }
  //
  while (1)
  {
    // stop this loop if message WOM_DONE has been detected
    if (this->callBckParam >= 2)
    {
      ::SetEvent(event);
      break;
    }
    ::Sleep(1);
  }   // end while (1)
  if (::logging)
    printf("WaveFile::play() callBckParam=%ld\n", this->callBckParam);
  //
  // Note: we don't close the WAVE device here even if it was opened here
  //     because it may be reused in other instances of this class
  //
  zurueck:
  if (::logging)
    printf("WaveFile::play() ret=%d\n", ret);
  return ret;
}   // end WaveFile::play()

// ---------------------------------------------------------
// This is a tokenizer for the file list in the command line
int getNextFn(char *strFn, char *actFn, int startPos)
{
  int pos  = startPos;
  int i    = 0;
  actFn[0] = 0;
  while (strFn[pos] != 0)
  {
    if (strFn[pos] == ' ')
    {
      pos++;
      continue;
    }
    if (strFn[pos] == '"')
    {
      pos++;
      continue;
    }
    if (strFn[pos] == 0)
      break;
    if (strFn[pos] == ',')
    {
      pos++;
      break;
    }
    actFn[i++] = strFn[pos];
    actFn[i]   = 0;
    pos++;
  }
  if (::strlen(actFn) == 0)
    pos = -1;
  return pos;
}   // end getNextFn()

// ---------------------------------------------------------
void usage()
{
  ::printf("usage: waveClass3 -i fileList [-v]\n");
}   // end usage()

// ---------------------------------------------------------
int main(int argc, char *argv[])
{
  int      ret = 0;
  int      cRet;
  DWORD    dwRet;
  int      startPos = 0;
  char     fileList[512] = "";
  char     *pActFn;
  ThreadParams *ptp = NULL;
  DWORD    dw_s, wt;
  DWORD    waitTime = 0L;
  HANDLE   s_t;
  HANDLE   evArray[MAX_NO_FILES];
  ThreadParams *threadParamsArray[MAX_NO_FILES];
  int      evArrayCnt = 0;
  //
  for (int n=0; n < MAX_NO_FILES; n++)
  {
    threadParamsArray[n] = NULL;
  }
  // the  filelist is a list of comma separated file names (WAVE files)
  for (int n=1; n < argc; n++)
  {
    if (::strcmp(argv[n], "-i") == 0)
      ::strcpy(fileList, argv[n + 1]); 
    if (::strcmp(argv[n], "-v") == 0)
      ::logging = 1;
  }
  if (::strlen(fileList) == 0)
  {
    usage();
    ret = -1;
    goto zurueck;
  }
  // create an event that should be signalled by the thread to indicate that 
  // the play process is finished
  //
  while (1)
  {
    pActFn = new char[256];
    startPos = getNextFn(fileList, pActFn, startPos);
    if (::logging)
      ::printf("main() next fn=%s\n", pActFn);
    if (startPos < 0)
      // no further file name found
      break;
    //
    // Create a separate thread for each file to be played
    // Associate an event with each thread
    // setup the  thread parameters
    ptp = new ThreadParams;
    threadParamsArray[evArrayCnt] = ptp;
    ptp->event = ::CreateEvent(0, false, false, 0);
    evArray[evArrayCnt++] = ptp->event;
    ::strcpy(ptp->actFn, pActFn);
    //
    // start the thread
    s_t = ::CreateThread(NULL,
                         0, 
                         workerRoutine, 
                         (LPVOID)ptp, 
                         0, 
                         &dw_s);
    ret++;
  }   // end while(1)
  //
  if (evArrayCnt > 0)
  {
    // wait for all events that are sent when the threads terminate
    dwRet = ::WaitForMultipleObjects((DWORD)evArrayCnt, 
                                     evArray,
                                     TRUE,
                                     INFINITE);
    ::Sleep(5);
    if (::logging)
      // dwRet = 0 means: event has been sent, 0x102=Timeout
      printf("main() result waiting on event=%ld\n", (long)dwRet);
  }
  //
  // everything OK here
  zurueck:
  for (int n=0; n < evArrayCnt; n++)
  {
    ptp = threadParamsArray[n];
    if (ptp != NULL)
      delete ptp;
  }
  ::Sleep(5);
  if (::logging)
    printf("main() Play over ret=%d\n", ret);
  return ret;
}   // end main()