// module wavdeco.c
// Read and decode a WAVE file
// Copyright Dr.E.Huckert 1-99
//
// versions:
// 17.04.2012 bug fixed (open()/read() not prepared for binary file)
//            bug fixed in read_data_chunk()
// 23.12.2015 fix for multiple data chunks where other chunks then 'data'
//                     preceed the data chunks (ex.: 'LIST')

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>

#define DWORD long
#define WORD short

// file size is typically for 1 second of samples:
//   44 bytes header
//  +   sample rate * bytes_per_sample
// ex.: 88244 bytes for a mono file at 44100 samples/second = CD quality
//
typedef struct
{
  DWORD main_chunk;  // "RIFF"
  DWORD mlength;     // Laenge ueber alles ohne die ersten 8 Bytes
  DWORD chunk_type;  // "WAVE"
} MAIN_CHUNK;        // 12 bytes

typedef struct       // start byte #12
{
  DWORD sub_chunk;   // muss 'fmt ' sein
  DWORD slength;     // #16
  WORD  format;      // #18
  WORD  channels;    // #20  1=mono, 2=stereo
  DWORD sample_fq;   // #24
  DWORD bytes_per_second;
  WORD  bytes_per_sample;
  WORD  bits_per_sample;
} SUB_CHUNK;        // 24 bytes

typedef struct      // start at pos. (12 + 24)=36
{
  DWORD data_type;   // kann 'data' sein oder 'LINK' oder ...
  DWORD dlength;
}  DATA_CHUNK;      // 8 bytes

// ----------------------------------------------------------
// read 12 bytes  main chunk header
// Returns: the number of bytes read
//          < 0 if error
int read_main_chunk(int ifd)
{
  int nb;
  char buf[20];
  char *caddr;
  MAIN_CHUNK header;
  int noRead = 0;
  //
  nb = read(ifd, &header, sizeof(MAIN_CHUNK));
  if (nb != sizeof(MAIN_CHUNK))
  {
    printf("ERROR: cannot read header of main chunk\n");
    return -1;
  }
  noRead += nb;
  memcpy(buf, (char *)(&(header.main_chunk)), 4);
  buf[4] = 0;
  printf("Kennung:     %08lx %s\n", header.main_chunk, buf);
  printf("Laenge:      %ld\n",      header.mlength);
  caddr = (char *)(&(header.main_chunk));
  memcpy(buf, caddr + 8, 4);   
  buf[4] = 0;
  printf("Typ:         %08lx %s\n",  header.chunk_type, buf);
  if (strncmp((char *)&(header.main_chunk), "RIFF", 4) != 0)
  {
     fprintf(stderr, "ERROR: wrong file tag (should be RIFF)\n");
     return -2; 
  }
  return noRead;
}   // end read_main_chunk()

// -------------------------------------------------------------
// read 24 bytes  sub chunk header
// starts at byte #12
// Returns: the number of bytes read
//          < 0 if error
int read_sub_chunk(int ifd)
{
  int nb;
  char buf[20];
  SUB_CHUNK header;
  int noRead = 0;
  //
  nb = read(ifd, &header, sizeof(SUB_CHUNK));
  if (nb != sizeof(SUB_CHUNK))
  {
    return -1;
  }
  noRead += nb;
  strncpy(buf, &(header.sub_chunk), 4);
  buf[4] = 0;
  printf("Sub-Kennung: %08lx %s\n", header.sub_chunk, buf); // " fmt"
  printf("Sub-Laenge:  %ld\n",  header.slength);  // 4 Bytes
  printf("Format:      %04x\n", header.format);   // 2 Bytes
  printf("Kanaele:     %04x\n", header.channels);    // 2 Bytes
  printf("Sample-Freq: %ld\n",  header.sample_fq);
  printf("Bytes/Sample:%d\n",   header.bytes_per_sample);
  printf("Bits/Sample: %d\n", header.bits_per_sample);
  if (strncmp((char *)&(header.sub_chunk), "fmt ", 4) != 0)
  {
     fprintf(stderr, "Falsche Unterkennung\n");
     return -1; 
  }
  return noRead;;
}   // end read_sub_chunk()

// ---------------------------------------------------------------
// read 8 bytes  data chunk header + all data bytes
// Returns: the number of bytes read
//          < 0 if error
long read_data_chunk(int ifd, 
                     DATA_CHUNK *pHeader)
{
  int        nb;
  long       ll;
  long       noRead = 0L;
  char       buf[512]; 
  static unsigned frameCount = 0;
  //
  // read 8 bytes = 2 DWORDs
  nb = read(ifd, pHeader, sizeof(DATA_CHUNK)); 
  if (nb <= 0)
  {
    return -1L;
  }
  noRead += nb;
  memcpy(buf, (char *)(&(pHeader->data_type)), 4);
  buf[4] = 0;
  frameCount++;
  printf("frame count=%u\n", frameCount);
  printf("Kennung: %08lx %s\n", pHeader->data_type, buf);
  printf("Laenge:  %ld\n", pHeader->dlength);
  /*
      if (strncmp((char *)&(pHeader->data_type), "data", 4) != 0)
      {
         fprintf(stderr, "Falscher Datenblock\n");
         return -2L; 
      }
      */
  // we read (consume) here the data - but we don't store them
  ll = pHeader->dlength;
  while (ll > 0L)
  {
    nb = sizeof(buf);
    if (ll < nb) nb = ll;
    nb = read(ifd, buf, nb);
    if (nb <= 0) 
    {
      break;
    }
    ll -= nb;
    noRead += nb;
  }
  //
  return noRead;
}   // end read_data_chunk()

// --------------------------------------------------------------
int main(int argc, char *argv[])
{
  int  ifd = -1;
  int  ret = 0;
  int  nb;
  long lnb;
  char buf[512];
  long readPos = 0L;
  unsigned noFrame = 0;
  DATA_CHUNK dataHeader;
  //
  if (argc != 2)
  {
    fprintf(stderr, "usage: wavdeco wav_file\n");
    ret = 0;
    goto zurueck;
  }
  ifd = open(argv[1], O_RDONLY + O_BINARY);
  if (ifd < 0)
  {
    ret = -1;
    goto zurueck;
  }
  printf("file:        %s\n", argv[1]);
  // ersten Header lesen: "RIFF", Laenge, "WAVE"
  nb = read_main_chunk(ifd);
  if (nb < 0)
  {
    fprintf(stderr, "main chunk error - we terminate!\n");
    ret = -2;
    goto zurueck;
  }
  readPos += nb;
  printf("after main chunk read pos=%ld 0x%lX\n", 
          readPos, readPos);
  printf("\n");
  // zweiten Header lesen (sub chunk, format chunk)
  nb = read_sub_chunk(ifd);
  if (nb < 0)
  {
    fprintf(stderr, "Fehler Dateiunterkennung\n");
    ret = -3;
    goto zurueck;
  }
  readPos += nb;
  printf("after format chunk read pos=%ld 0x%lX\n", 
          readPos, readPos);
  printf("\n");
  // 
  // Eine Folge von DATA_CHUNK frames lesen
  // Die eigentlichen Data stehen in chunks vom Typ 'data'
  while (1)
  {
    lnb = read_data_chunk(ifd, &dataHeader);
    if (lnb <= 0)
      break;
    readPos += lnb;
    noFrame++;
    printf("after frame=%u read pos=%ld 0x%lX\n", 
           noFrame, readPos, readPos);
    printf("\n");
  }  // end while 1
  //
  ret = 1;
  //
  zurueck:
  printf("last read pos=%ld\n", readPos);
  printf("return code=%d\n", ret);
  if (ifd >= 0)
    close(ifd);
  return ret;
}   // end main()
